class Encryptor:
    def __init__(self, public_key1, public_key2, private_key):
        self.public_key1 = public_key1
        self.public_key2 = public_key2
        self.private_key = private_key
        self.full_key = None

    def generate_partial_key(self):
        partial_key = self.public_key1 ** self.private_key
        partial_key = partial_key % self.public_key2
        return partial_key

    def generate_full_key(self, partial_key_r):
        full_key = partial_key_r ** self.private_key
        full_key = full_key % self.public_key2
        self.full_key = full_key
        return full_key

    def encrypt_message(self, message):
        encrypted_message = ""
        key = self.full_key
        for c in message:
           encrypted_message += chr(ord(c) + key)
        return encrypted_message

    def decrypt_message(self, encrypted_message):
        decrypted_message = ""
        key = self.full_key
        for c in encrypted_message:
           decrypted_message += chr(ord(c) - key)
        return decrypted_message



from abc import ABC, abstractmethod


#  Source code: https://docs.python.org/3/howto/descriptor.html#complete-practical-example


class Validator(ABC):
    @abstractmethod
    def validate(self, value):
        pass


class OneOf(Validator):
    def __init__(self, *options):
        self.options = set(options)

    def validate(self, value):
        if value not in self.options:
            raise ValueError(f'Expected {value!r} to be one of {self.options!r}')
        return value


class Number(Validator):
    def __init__(self, minvalue: int = None, maxvalue: int = None):
        self.minvalue = minvalue
        self.maxvalue = maxvalue

    def validate(self, value):
        try:
            value = int(value)
        except ValueError:
            raise TypeError(f'Expected {value!r} to be an int or float')
        if self.minvalue is not None and value < self.minvalue:
            raise ValueError(
                f'Expected {value!r} to be at least {self.minvalue!r}'
            )
        if self.maxvalue is not None and value > self.maxvalue:
            raise ValueError(
                f'Expected {value!r} to be no more than {self.maxvalue!r}'
            )
        return value


class String(Validator):
    def __init__(self, minsize=None, maxsize=None, predicate=None):
        self.minsize = minsize
        self.maxsize = maxsize
        self.predicate = predicate

    def validate(self, value):
        if not isinstance(value, str):
            raise TypeError(f'Expected {value!r} to be an str')
        if self.minsize is not None and len(value) < self.minsize:
            raise ValueError(
                f'Expected {value!r} to be no smaller than {self.minsize!r}'
            )
        if self.maxsize is not None and len(value) > self.maxsize:
            raise ValueError(
                f'Expected {value!r} to be no bigger than {self.maxsize!r}'
            )
        if self.predicate is not None and not self.predicate(value):
            raise ValueError(
                f'Expected {self.predicate} to be true for {value!r}'
            )
        return value



import functools
import random
from asyncio import StreamReader, StreamWriter

import asyncio

from hasbulla_boom.database import db
from hasbulla_boom.encryption import Encryptor


class Server:
    def __init__(self, host: str = "localhost", port: int = 8888):
        self.host = host
        self.port = port
        self.users = []

    @staticmethod
    def register_command(*types):
        len_types = len(types)

        def decorator(func):
            @functools.wraps(func)
            def wrapper(user_id: int, *args, **kwargs):
                assert len_types == len(args) + len(kwargs) + 1, f'Get command takes exactly {len_types - 1} arguments'
                for i in range(1, len_types):
                    assert types[i].validate(args[i - 1])
                return func(user_id, *args, **kwargs)
            db.commands_dict[func.__name__] = wrapper
            return wrapper
        return decorator

    @staticmethod
    def handle_command(user_id: int, cmd: str, *args) -> str:
        try:
            command_func = db.commands_dict.get(cmd)
            if command_func is None:
                raise ValueError('Unknown command')
            result = command_func(user_id, *args)
        except (TypeError, ValueError, AssertionError) as e:
            result = str(e)
        except StopIteration:
            result = ''
        return result

    @staticmethod
    async def keys_exchange(reader: StreamReader, writer: StreamWriter) -> Encryptor:
        public_key = random.randint(1000, 10000)
        private_key = random.randint(1000, 10000)
        writer.write((str(public_key) + "\n").encode('utf8'))
        await writer.drain()
        client_public_key = int((await reader.readline()).decode('utf8'))
        encryptor = Encryptor(public_key, client_public_key, private_key)
        writer.write((str(encryptor.generate_partial_key()) + "\n").encode('utf8'))
        await writer.drain()
        client_partial_key = int((await reader.readline()).decode('utf8'))
        encryptor.generate_full_key(client_partial_key)
        return encryptor

    @staticmethod
    async def get_notifications(user_id: int, writer: StreamWriter, encryptor: Encryptor) -> None:
        if db.notifications[user_id]:
            response = "\n".join([i for i in db.notifications[user_id]])
            db.notifications[user_id] = []
            writer.write((encryptor.encrypt_message(response) + "\n").encode('utf8'))
            await writer.drain()

    async def handle_client(self, reader: StreamReader, writer: StreamWriter) -> None:
        user_id = len(self.users)
        self.users.append(user_id)
        db.notifications[user_id] = []
        encryptor = await self.keys_exchange(reader, writer)
        writer.write((encryptor.encrypt_message("Hi client") + "\n").encode('utf8'))
        await writer.drain()
        request = encryptor.decrypt_message((await reader.readline()).decode('utf8').rstrip('\n'))
        while request:
            result = self.handle_command(user_id, *request.split())
            if result == "stop":
                break
            await self.get_notifications(user_id, writer, encryptor)
            response = encryptor.encrypt_message(result) + '\n'
            writer.write(response.encode('utf8'))
            await writer.drain()
            await self.get_notifications(user_id, writer, encryptor)
            request = encryptor.decrypt_message((await reader.readline()).decode('utf8').rstrip('\n'))

        writer.close()

    async def run(self) -> None:
        server = await asyncio.start_server(self.handle_client, self.host, self.port)

        async with server:
            await server.serve_forever()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hasbulla_boom']

package_data = \
{'': ['*']}

install_requires = \
['aiomisc>=15.6.8,<16.0.0',
 'asyncio>=3.4.3,<4.0.0',
 'pytest-asyncio>=0.18.1,<0.19.0',
 'pytest-cov>=3.0.0,<4.0.0',
 'pytest>=7.0.1,<8.0.0']

setup_kwargs = {
    'name': 'hasbulla-boom',
    'version': '3.2.2',
    'description': '',
    'long_description': None,
    'author': 'Your Name',
    'author_email': 'you@example.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
